/*
    gvidm - X11 video mode changer with a minimal interface
    Copyright (C) 2001-2003 Matthew Mueller <donut@dakotacom.net>
    based on gvid:
    Copyright (C) 1999-2001 Keith Vanderline <kvand@mit.edu>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "gvid.h"
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif

#include <vector>
#include <string>
#include "_ios.h"
#include "_sstream.h"
#include <iomanip>

using namespace std;

typedef vector<int> mode_list_t;
typedef vector<int> screen_list_t;
typedef vector<mode_list_t> screen_mode_list_t;


Display *display;

int switch_X11_Res(Display *display, int screen, int mode)
{
	XF86VidModeModeInfo **vm_modelines;
	int vm_count;

	XF86VidModeGetAllModeLines(display, screen, &vm_count, &vm_modelines);
	assert(mode<vm_count);
	XF86VidModeSwitchToMode(display, screen, vm_modelines[mode]);
	XFree(vm_modelines);
	XFlush(display);
	return 1;
}

int get_screen_count(void) {
#ifdef HAVE_LIBXINERAMA
	if (XineramaIsActive(display))
	{
		int screen_exists[1024], i;
		XineramaScreenInfo *screen = XineramaQueryScreens(display, screen_exists);
		for (i=0; i==screen[i].screen_number; i++){}
		XFree(screen);
		return i;
	}
	else
	{
#endif
		return XScreenCount(display);
#ifdef HAVE_LIBXINERAMA
	}
#endif
}

void MenuEvent (GtkWidget *widget, gpointer mode_ptr)
{
	int mode_int, screen;

	memcpy(&mode_int, mode_ptr, sizeof(mode_int));
	screen = mode_int >> 16;
	mode_int = mode_int & 0xFFFF;
	assert(screen<get_screen_count());

	switch_X11_Res(display, screen, mode_int);
}

void on_deactivate(void) {
	gtk_main_quit();
}

GtkWidget *create_menuitem_with_accel(const char *label_str, GtkWidget *menu_ptr) {
	GtkWidget *menuitem_ptr = gtk_menu_item_new_with_label("");
	guint tmp_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (menuitem_ptr)->child), label_str);
	if (tmp_key!=GDK_VoidSymbol) {
		GtkAccelGroup *menu_accels = gtk_menu_ensure_uline_accel_group (GTK_MENU (menu_ptr));
		gtk_widget_add_accelerator (menuitem_ptr, "activate_item", menu_accels, tmp_key, 0, (GtkAccelFlags)0);
	}
	return menuitem_ptr;
}

inline char num2char36(int i) {
	if (i<10) return '0'+i;
	return 'a'+i-10;
}

inline bool is_same_mode(XF86VidModeModeInfo *mi, int dotclock, XF86VidModeModeLine *ml)
{
	return mi->dotclock == (unsigned int)dotclock &&
		mi->hdisplay == ml->hdisplay && mi->vdisplay == ml->vdisplay &&
		mi->htotal == ml->htotal     && mi->vtotal == ml->vtotal &&
		mi->flags == ml->flags;
}

inline float calc_refresh(int dotclock, short htotal, short vtotal, unsigned int flags) {
	return (dotclock * 1000.0 / (htotal * vtotal) * ((flags & XF86CONF_DBLSCAN)?0.5:1));
}

inline float calc_refresh(const XF86VidModeModeInfo *m) {
	return calc_refresh(m->dotclock, m->htotal, m->vtotal, m->flags);
}

inline float calc_refresh(int dotclock, const XF86VidModeModeLine *m) {	
	return calc_refresh(dotclock, m->htotal, m->vtotal, m->flags); 
}

string modestring(short hdisplay, short vdisplay, float refresh, unsigned int flags) {
	ostringstream oss;
	oss << hdisplay << 'x' << vdisplay << '@' << fixed << setprecision(0) << refresh;
	return oss.str();
}

string pretty_modestring(short hdisplay, short vdisplay, float refresh, unsigned int flags) {	
	ostringstream oss;
	oss << setw(4) << hdisplay << 'x' << left << setw(4) << vdisplay << ' ' << fixed << setprecision(0) << refresh << "Hz";
	if (flags & (XF86CONF_DBLSCAN|XF86CONF_INTERLACE)) {
		oss << " (";
		if (flags & XF86CONF_DBLSCAN) oss << 'd';
		if (flags & XF86CONF_INTERLACE) oss << 'i';
		oss << ')';
	}
	return oss.str();
}

int round(float f) {
	//This is a cheesy yet effective workaround for the fact that
	//round/nearbyint/etc are all C99.  (And rint which is bsd.)
	//On the plus side, it guarantees that no matter which rounding method
	//iostreams use, user input will always match up to the output of --list.
	return atoi(strchr(modestring(0,0,f,0).c_str(), '@')+1);
}

inline string menuline(bool is_curmode, int i, const XF86VidModeModeInfo *m)
{
	ostringstream oss;
	oss << (is_curmode?'*':' ');
	if (i < 36) oss << " _" << num2char36(i) << ". ";
	else oss << "    ";
	oss << pretty_modestring(m->hdisplay, m->vdisplay, calc_refresh(m), m->flags);
	return oss.str();
}

void fill_res_menu(GtkWidget *menu_ptr, int screen, const mode_list_t &modes) {
	int mode, vm_count, *mode_ptr, dotclock, i=0;
	XF86VidModeModeInfo **vm_modelines;
	XF86VidModeModeLine modeline;
	GtkWidget *menuitem_ptr;

	XF86VidModeGetAllModeLines(display, screen, &vm_count, &vm_modelines);
	XF86VidModeGetModeLine(display, screen, &dotclock, &modeline);
	for (mode_list_t::const_iterator modeitr=modes.begin(); modeitr!=modes.end(); ++modeitr, ++i)
	{
		mode = *modeitr;
		assert(mode<vm_count);
		string s = menuline(is_same_mode(vm_modelines[mode], dotclock, &modeline), i, vm_modelines[mode]);

		mode_ptr = (int*)malloc(sizeof(int));
		*mode_ptr = (screen<<16) + mode;

		menuitem_ptr = create_menuitem_with_accel(s.c_str(), menu_ptr);
		gtk_signal_connect(GTK_OBJECT (menuitem_ptr), "activate", GTK_SIGNAL_FUNC (MenuEvent), (gpointer)mode_ptr);
		gtk_menu_append(GTK_MENU (menu_ptr), menuitem_ptr);
		gtk_widget_show(menuitem_ptr);
		gtk_widget_lock_accelerators(menuitem_ptr);
	}
	XFree(vm_modelines);
}

void create_popup(const screen_mode_list_t &screen_modes)
{
	int i, screen_count, mode_number;
	char label_str[80];
	GtkWidget *menuitem_ptr, *submenu_ptr, *menu_ptr;

	mode_number=0;
	screen_count = get_screen_count();

	menu_ptr = gtk_menu_new();
	gtk_signal_connect (GTK_OBJECT (menu_ptr), "deactivate", GTK_SIGNAL_FUNC (on_deactivate), NULL);

	if (screen_count==1)
	{
		assert(!screen_modes[0].empty());
		fill_res_menu(menu_ptr, 0, screen_modes[0]);
	}
	else
	{
		for (i=0; i<screen_count; i++)
		{
			if (screen_modes[i].empty())
				continue;
			
			submenu_ptr = gtk_menu_new();
			sprintf(label_str, "Screen %s%i", i<10?"_":"", i);
			menuitem_ptr = create_menuitem_with_accel(label_str, menu_ptr);
			gtk_menu_item_set_submenu(GTK_MENU_ITEM (menuitem_ptr), submenu_ptr);
			gtk_menu_append(GTK_MENU (menu_ptr), menuitem_ptr);
			gtk_widget_show(menuitem_ptr);
			gtk_widget_show(submenu_ptr);

			fill_res_menu(submenu_ptr, i, screen_modes[i]);
		}
	}

	gtk_menu_popup(GTK_MENU(menu_ptr), NULL, NULL, NULL, NULL, 0, 0);
}

int parse_int(const string &arg, int &i) {
	char *endptr;
	i = strtol(arg.c_str(), &endptr, 10);
	if (arg.empty() || *endptr!='\0') return 1;
	return 0;
}

int parse_screen(const string &arg, screen_list_t &screens) {
	if (arg == "all") {
		for (int screen=0; screen < get_screen_count(); ++screen)
			screens.push_back(screen);
	} else {
		int screen;
		if (parse_int(arg, screen)) return 1;
		if (screen<0 || screen>=get_screen_count()) return 1;
		screens.push_back(screen);
	}
	return 0;
}

int parse_mode(const string &arg, int &w, int &h, int &r) {
	r=-1;
	if (arg == "max")
		w = h = INT_MAX;
	else if (arg == "all")
		w = h = -1;
	else {	
		int x = arg.find('x');
		int a = arg.find('@');
		int hlen = ((a>=0)?a:arg.size())-x-1;
		if (x<0) return 1;
		if (parse_int(arg.substr(0,x),w) || parse_int(arg.substr(x+1,hlen),h)) return 1;
		if (a>=0 && parse_int(arg.substr(a+1), r)) return 1;
		if (w<=0 || h<=0) return 1;
	}
	return 0;
}

int parse_arg(const string &arg, screen_list_t &screens, int &w, int &h, int &r) {
	int c = arg.find(':');
	if (c>=0) {
		if (parse_screen(arg.substr(0, c), screens)) return 1;
		if (parse_mode(arg.substr(c+1), w, h, r)) return 1;
	}
	else {
		screens.push_back(0);
		if (parse_mode(arg, w, h, r)) return 1;
	}
	return 0;	
}

int find_mode(int screen, int w, int h, int r, mode_list_t &modes) {
	XF86VidModeModeInfo **vm_modelines;
	int vm_count, j, errs=0;

	XF86VidModeGetAllModeLines(display, screen, &vm_count, &vm_modelines);
	if (w==INT_MAX && h==INT_MAX) {
		unsigned long maxpixels = 0;
		int max_mode = 0;
		for (j=0; j<vm_count; j++) {
			unsigned long pixelcount = vm_modelines[j]->hdisplay * vm_modelines[j]->vdisplay;
			if (pixelcount > maxpixels) {
				maxpixels = pixelcount;
				max_mode = j;
			}
		}
		modes.push_back(max_mode);
	}
	else if (w==-1 && h==-1) {
		for (j=0; j<vm_count; ++j)
			modes.push_back(j);
	}
	else {
		for (j=0; j<vm_count; j++)
			if (vm_modelines[j]->hdisplay==w && vm_modelines[j]->vdisplay==h) {
				if (r>=0) {
					float refresh = calc_refresh(vm_modelines[j]);
					if (round(refresh) != r)
						continue;
				}
				modes.push_back(j);
				break;
			}
		if (j>=vm_count) errs++;
	}
	XFree(vm_modelines);
	return errs;
}

int parse_arg_mode(screen_mode_list_t &arg_modes, const char *arg) {
	screen_list_t screens;
	int w, h, r, errs = 0;

	if (parse_arg(arg, screens, w, h, r)) {
		fprintf(stderr, "invalid mode \"%s\"\n",arg);
		errs++;
	}
	else {
		for (screen_list_t::const_iterator screenitr=screens.begin(); screenitr!=screens.end(); ++screenitr) {
			int screen = *screenitr;
			if (find_mode(screen, w, h, r, arg_modes[screen])) {
				fprintf(stderr, "screen %i mode %ix%i", screen, w, h);
				if (r>=0) fprintf(stderr, "@%i", r);
				fprintf(stderr, " not found\n");
				errs++;
			}
		}
	}
	return errs;
}

int parse_arg_modes(screen_mode_list_t &arg_modes, char **arg0, char **lastarg) {
	int errs = 0;
	for (; arg0<=lastarg; ++arg0) {
		errs += parse_arg_mode(arg_modes, *arg0);
	}
	return errs;
}

int parse_arg_screen(screen_list_t &screens, const char *arg) {
	if (parse_screen(arg, screens)) {
		fprintf(stderr, "invalid screen \"%s\"\n", arg);
		return 1;
	}
	return 0;
}

int parse_arg_screens(screen_list_t &screens, char **arg0, char **lastarg) {
	int errs = 0;
	for (; arg0<=lastarg; ++arg0) {
		errs += parse_arg_screen(screens, *arg0);
	}
	return errs;
}

void list_modes(const screen_list_t &screens) {
	int num_screens=get_screen_count();
	for (screen_list_t::const_iterator screenitr=screens.begin(); screenitr!=screens.end(); ++screenitr) {
		int screen = *screenitr;
		int vm_count, dotclock;
		XF86VidModeModeLine modeline;
		XF86VidModeGetModeLine(display, screen, &dotclock, &modeline);
		XF86VidModeModeInfo **vm_modelines;
		XF86VidModeGetAllModeLines(display, screen, &vm_count, &vm_modelines);
		for (int mode=0; mode < vm_count; mode++) {
			XF86VidModeModeInfo *m = vm_modelines[mode];
			if (num_screens > 1) printf("%i:", screen);
			printf("%s%s\n", modestring(m->hdisplay, m->vdisplay, calc_refresh(m), m->flags).c_str(),
				is_same_mode(m, dotclock, &modeline)?" *":"");
		}
		XFree(vm_modelines);
	}
}

void list_current_modes(const screen_list_t &screens) {
	int num_screens=get_screen_count();
	for (screen_list_t::const_iterator screenitr=screens.begin(); screenitr!=screens.end(); ++screenitr) {

		int screen = *screenitr;
		int dotclock;
		XF86VidModeModeLine modeline;
		XF86VidModeGetModeLine(display, screen, &dotclock, &modeline);
		if (num_screens > 1) printf("%i:", screen);
		printf("%s\n", modestring(modeline.hdisplay, modeline.vdisplay, calc_refresh(dotclock, &modeline), modeline.flags).c_str());
	}
}

int check_VidMode_supported(void) {
	int major,minor;
	return !XF86VidModeQueryVersion(display, &major, &minor);
}

void show_help_and_exit(int code) {
	printf("gvidm "M_VERSION" (C) 2001-2003 Matthew Mueller <donut@dakotacom.net>\n"
			"based on gvid "VERSION" (C) 1999-2001 Keith Vanderline <kvand@mit.edu>\n"
			"Distributed under the terms of the GPL\n");
	printf("USAGE: gvidm [modes...]\n"
			"  -h/--help   show this info\n"
			"  -q/--query  print current resolutions\n"
			"  -l/--list   print list of available resolutions\n"
			"\n"
			" Modes are specified as [screen:]<resolution>\n"
			"   screen may be a number or the keyword \"all\".\n"
			"   resolution may be <width>x<height>[@hz] or the keyword \"max\" or \"all\"\n"
			" If no modes are specified, a menu of all available resolutions is shown.\n"
			" If one mode(per screen) is specified, it is switched to immediatly.\n"
			" If multiple modes are specified, a menu with only those modes is shown.\n"
	);
	exit(code);
}

int main(int argc,char **argv)
{
#ifdef HAVE_GETOPT_LONG
	static struct option long_options[] = {
		{"help", 0, 0, 'h'},
		{"list", 0, 0, 'l'},
		{"query", 0, 0, 'q'},
		{0, 0, 0 ,0}
	};
#else
#define getopt_long(argc,argv,shopt,lopt,lidx) getopt(argc,argv,shopt)
#endif
	int c;
	enum {
		MODE_SET,
		MODE_LIST,
		MODE_LIST_CURRENT,
	} mode = MODE_SET;

	if ((display=XOpenDisplay(""))==NULL) {
		fprintf(stderr, "gvidm: Error opening display\n");
		exit(1);
	}
	if (check_VidMode_supported()) {
		fprintf(stderr, "gvidm: VidMode extension not supported\n");
		exit(1);
	}

	gtk_init (&argc, &argv);
	while (-1 != (c=getopt_long(argc, argv, "hlq", long_options, NULL))) {
		switch (c) {
			case ':':
			case '?':
				show_help_and_exit(1);
				break;
			case 'l':
				mode = MODE_LIST;
				break;
			case 'q':
				mode = MODE_LIST_CURRENT;
				break;
			case 'h':
				show_help_and_exit(0);
				break;
			default:
				assert(0);
		}
	}

	int errs=0;

	if (mode==MODE_LIST || mode==MODE_LIST_CURRENT) {
		screen_list_t screens;
		if (optind < argc) {
			errs += parse_arg_screens(screens, argv+optind, argv+argc-1);
		} else
			errs += parse_screen("all", screens);
		if (mode==MODE_LIST)
			list_modes(screens);
		else if (mode==MODE_LIST_CURRENT)
			list_current_modes(screens);
		return errs?1:0;
	}

	screen_mode_list_t arg_modes(get_screen_count());
	if (optind < argc) {
		int want_menu = 0;
		errs += parse_arg_modes(arg_modes, argv+optind, argv+argc-1);
		for (unsigned screen = 0; screen<arg_modes.size(); ++screen) {
			if (arg_modes[screen].empty())
				continue;
			else if (arg_modes[screen].size()==1) {
				switch_X11_Res(display, screen, arg_modes[screen].front());
				arg_modes[screen].clear(); // clear it so it won't display in the popup
			}
			else
				want_menu = 1;
		}
		if (!want_menu)
			return errs?1:0;
		create_popup(arg_modes);
	}
	else {
		parse_arg_mode(arg_modes, "all:all");
		create_popup(arg_modes);
	}
	gtk_main ();
	return errs?1:0;
}
